!for LineSearch
subroutine CalcEnergy(Pos, M, L, rc, PEnergy, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    real(8), intent(in) :: L, rc
    real(8), intent(out) :: PEnergy
    real(8), dimension(Dim) :: rij, Posi
    real(8) :: d2, id2, id6, id12, dL, rc2, Shift
    integer :: i, j
    PEnergy = 0.
    Shift = -4. * (rc**(-12) - rc**(-6))
    rc2 = rc * rc

    do i = 0, NAtom - 1
    
        Posi = Pos(i,:)
        
        do j = i+1, NAtom - 1
        
            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
            
            id2 = 1. / d2 !inverse squared distance
            id6 = id2 * id2 * id2 !inverse sixth distance
            id12 = id6 * id6 !inverse twelvth distance
            PEnergy = PEnergy + 4. * (id12 - id6) + Shift
            
        end do
    end do
    
end subroutine

!for CG method
subroutine CalcEnergyForces(Pos, M, L, rc, PEnergy, Forces, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    real(8), intent(in) :: L, rc
    real(8), intent(out) :: PEnergy
    real(8), dimension(Dim) :: rij, Fij, Posi
    real(8), intent(inout), dimension(0:NAtom-1, 0:Dim-1) :: Forces
!f2py intent(in, out) :: Forces
    real(8) :: d2, id2, id6, id12, dL, rc2, Shift
    integer :: i, j
    
    PEnergy = 0.
    Forces = 0.
    Shift = -4. * (rc**(-12) - rc**(-6))
    rc2 = rc * rc
    do i = 0, NAtom - 1
    
        Posi = Pos(i,:)
        
        do j = i+1, NAtom - 1
        
            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
            
            id2 = 1. / d2 !inverse squared distance
            id6 = id2 * id2 * id2 !inverse sixth distance
            id12 = id6 * id6 !inverse twelvth distance
            PEnergy = PEnergy + 4. * (id12 - id6) + Shift
            Fij = rij * ((-48. * id12 + 24. * id6) * id2)
            Forces(i,:) = Forces(i,:) + Fij
            Forces(j,:) = Forces(j,:) - Fij
            
        end do
    end do

end subroutine


!for lmbda ensemble
subroutine LmbdaLJCalcEnergy(lmbLJ, Pos, M, L, rc, PEnergy, dUdlmb, Dim, NAtom)
    implicit none
    integer, intent(in) :: M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    real(8), intent(in) :: L, rc, lmbLJ
    real(8), intent(out) :: PEnergy, dUdlmb
    real(8), dimension(Dim) :: rij, Posi
    real(8) :: d2, id2, id6, id12, rc2, Shift, lmbShift, lmb4, d6, reff, ireff, ireff2, reffrc, ireffrc, ireffrc2
    real(8) :: LJSoft, lmbPEnergy, alpha, alphalmb, alphalmb2, dUdlmbShift
    integer :: i, j
    PEnergy = 0.
    lmbPEnergy = 0.
    LJSoft = 0.
    dUdlmb = 0.
    alpha = 0.5
    alphalmb = alpha * (1-lmbLJ) 
    alphalmb2 = alphalmb * (1-lmbLJ)
    lmb4 = 4. * lmbLJ
    Shift = -4. * (rc**(-12) - rc**(-6))
    reffrc = alphalmb2 + (rc**6)
    ireffrc = 1./reffrc
    ireffrc2 = ireffrc * ireffrc
    lmbShift = -lmb4 * (ireffrc2 - ireffrc)
    dUdlmbShift = (lmbShift/lmbLJ) - (lmb4 * alphalmb * ireffrc * 2) * ( (2*ireffrc2) - ireffrc )
    rc2 = rc * rc

    i=0
    
    Posi = Pos(i,:)
        
    do j = i+1, NAtom - 1
    
        rij = Pos(j,:) - Posi
        rij = rij - L * dnint(rij/L)
        d2 = sum(rij * rij)
        
        if (d2>rc2) then !this is placed later on intentionally...
            cycle
        end if
        

        d6 = d2 * d2 * d2 !sixth distance
        reff = alphalmb2 + d6
        ireff = 1./reff
        ireff2 = ireff*ireff
        LJSoft = lmb4 * (ireff2 - ireff) 
        lmbPEnergy = lmbPEnergy + LJSoft + lmbShift
        
        dUdlmb = dUdlmb + (LJSoft/lmbLJ) + (lmb4 * alphalmb * ireff * 2) * ( (2*ireff2) - ireff ) + dUdlmbShift
    end do


    do i = 1, NAtom - 1
    
        Posi = Pos(i,:)
        
        do j = i+1, NAtom - 1
        
            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
            
            id2 = 1. / d2 !inverse squared distance
            id6 = id2 * id2 * id2 !inverse sixth distance
            id12 = id6 * id6 !inverse twelvth distance
            PEnergy = PEnergy + 4. * (id12 - id6) + Shift
            
        end do
    end do
    
    PEnergy = PEnergy + lmbPEnergy
    
end subroutine


subroutine qCalcEnergy(q, Pos, M, L, rc, PEnergy, Dim, NAtom)
    implicit none
    integer, intent(in) :: q, M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    real(8), intent(in) :: L, rc
    real(8), intent(out) :: PEnergy
    real(8), dimension(Dim) :: rij, Posi
    real(8) :: d2, id2, id6, id12, dL, rc2, Shift
    integer :: i, j
    PEnergy = 0.
    Shift = -4. * (rc**(-12) - rc**(-6))
    rc2 = rc * rc

    do i = 0, NAtom - 1
    
        Posi = Pos(i,:)
        
        do j = i+1, NAtom - 1
            if (i==q .or. j==q) then
                rij = Pos(j,:) - Posi
                rij = rij - L * dnint(rij/L)
                d2 = sum(rij * rij)
                
                if (d2>rc2) then !this is placed later on intentionally...
                    cycle
                end if
                
                id2 = 1. / d2 !inverse squared distance
                id6 = id2 * id2 * id2 !inverse sixth distance
                id12 = id6 * id6 !inverse twelvth distance
                PEnergy = PEnergy + 4. * (id12 - id6) + Shift
            end if
        end do
    end do


end subroutine


subroutine qlmbdaLJCalcEnergy(q, lmbLJ, Pos, M, L, rc, PEnergy, dUdlmb, Dim, NAtom)
    implicit none
    integer, intent(in) :: q, M, Dim, NAtom
    real(8), intent(in), dimension(0:NAtom-1, 0:Dim-1) :: Pos
    real(8), intent(in) :: L, rc, lmbLJ
    real(8), intent(out) :: PEnergy, dUdlmb
    real(8), dimension(Dim) :: rij, Posi
    real(8) :: d2, id2, id6, id12, rc2, Shift, lmbShift, lmb4, d6, reff, ireff, ireff2, reffrc, ireffrc, ireffrc2
    real(8) :: LJSoft, lmbPEnergy, alpha, alphalmb, alphalmb2, dUdlmbShift
    integer :: i, j
    PEnergy = 0.
    lmbPEnergy = 0.
    LJSoft = 0.
    dUdlmb = 0.
    alpha = 0.5
    alphalmb = alpha * (1-lmbLJ) 
    alphalmb2 = alphalmb * (1-lmbLJ)
    lmb4 = 4. * lmbLJ
    Shift = -4. * (rc**(-12) - rc**(-6))
    reffrc = alphalmb2 + (rc**6)
    ireffrc = 1./reffrc
    ireffrc2 = ireffrc * ireffrc
    lmbShift = -lmb4 * (ireffrc2 - ireffrc)
    dUdlmbShift = (lmbShift/lmbLJ) - (lmb4 * alphalmb * ireffrc * 2) * ( (2*ireffrc2) - ireffrc )
    rc2 = rc * rc

    i=q
    
    Posi = Pos(i,:)

    if (q == 0) then
    
        do j = i+1, NAtom - 1

            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
            
            d6 = d2 * d2 * d2 !sixth distance
            reff = alphalmb2 + d6
            ireff = 1./reff
            ireff2 = ireff*ireff
            LJSoft = lmb4 * (ireff2 - ireff) 
            lmbPEnergy = lmbPEnergy + LJSoft + lmbShift
        
            dUdlmb = dUdlmb + (LJSoft/lmbLJ) + (lmb4 * alphalmb * ireff * 2) * ( (2*ireff2) - ireff ) + dUdlmbShift
            
        end do
        
    else
    
        j=0
        rij = Pos(j,:) - Posi
        rij = rij - L * dnint(rij/L)
        d2 = sum(rij * rij)   
          
        if (d2<rc2) then !this is placed later on intentionally...
            d6 = d2 * d2 * d2 !sixth distance
            reff = alphalmb2 + d6
            ireff = 1./reff
            ireff2 = ireff*ireff
            LJSoft = lmb4 * (ireff2 - ireff) 
            lmbPEnergy = lmbPEnergy + LJSoft + lmbShift
        
            dUdlmb = dUdlmb + (LJSoft/lmbLJ) + (lmb4 * alphalmb * ireff * 2) * ( (2*ireff2) - ireff ) + dUdlmbShift
        end if

                           
        do j = 1, q - 1
    
            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
          
            id2 = 1. / d2 !inverse squared distance
            id6 = id2 * id2 * id2 !inverse sixth distance
            id12 = id6 * id6 !inverse twelvth distance
            PEnergy = PEnergy + 4. * (id12 - id6) + Shift
    
        end do
    
        do j = q+1, NAtom - 1
    
            rij = Pos(j,:) - Posi
            rij = rij - L * dnint(rij/L)
            d2 = sum(rij * rij)
            
            if (d2>rc2) then !this is placed later on intentionally...
                cycle
            end if
            
            id2 = 1. / d2 !inverse squared distance
            id6 = id2 * id2 * id2 !inverse sixth distance
            id12 = id6 * id6 !inverse twelvth distance
            PEnergy = PEnergy + 4. * (id12 - id6) + Shift
    
        end do
    end if
    
    PEnergy = PEnergy + lmbPEnergy
    
end subroutine